@@ -2,7 +2,8 @@
 from __future__ import unicode_literals
 
 import os
-
+import sys
+sys.path.insert(0, os.path.abspath('../src'))
 
 extensions = [
     'sphinx.ext.autodoc',
@@ -23,10 +24,10 @@ if os.getenv('SPELLCHECK'):
 source_suffix = '.rst'
 master_doc = 'index'
 project = 'Python-ASN1'
-year = '2007-2016'
+year = '2007-2022'
 author = 'Sebastien Andrivet'
 copyright = '{0}, {1}'.format(year, author)
-version = release = '2.2.0'
+version = release = '2.5.0'
 
 pygments_style = 'trac'
 templates_path = ['.']
